//
// "$Id ToolGroup.cxx Thu Feb 7 2006 z_hossain $"
//
// Dockability for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2005 by Bill Spitzak and others.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems to: 
//
//             md_z_hossain@yahoo.com
//

#include <fltk/events.h>
#include <fltk/Box.h>
#include <fltk/Item.h>
#include <fltk/Divider.h>
#include <fltk/draw.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "ToolGroup.h"
#include "ToolDivider.h"
#include "DockRegion.h"

#define ITEM_GAP 2

#define VERTICAL_MENUINDEX 10
#define HORIZONTAL_MENUINDEX 11
#define FLOAT_MENUINDEX 12
#define HIDE_MENUINDEX 13

using namespace fltk;

struct MenuIndex {
	int index_;
	tgGripper *gripper_;
	
	MenuIndex(int i, tgGripper *g) {
		index_ = i;
		gripper_ = g;
	}
};

ToolDivider::ToolDivider(int g) : Widget(0, 0, g, g) {
	style(Item::default_style);
	set_flag(ALIGN_LEFT|ALIGN_INSIDE|OUTPUT);
	gap_ = g;
}

void ToolDivider::layout() {
	if(h() > w()) w(gap_);
	else h(gap_);
}

void ToolDivider::draw()
{
	draw_background();
	setcolor( GRAY33 );
	if ( w() > h() ) {
		int y = ( h() - 1 ) / 2;
		drawline(4, y, w() - 5, y);
		setcolor(WHITE);
		drawline(4, y + 1, w() - 5, y + 1);
	}
	else if ( h() ) {
		int x = (w() - 1) / 2;
		drawline(x, 4, x, h() - 5);
		setcolor(WHITE);
		drawline(x + 1, 4, x + 1, h() - 5);
	}
}

tgGripper::tgGripper(int t) : PopupMenu(0, 0, 10, 40) {
	thick_ = t;
	box(UP_BOX);
	begin();
	Item* o;
	o = new Item("Dock &Top");
	o->callback(menu_cb, new MenuIndex(DockableGroup::DOCKED_TOP, this));
	o = new Item("Dock &Bottom");
	o->callback(menu_cb, new MenuIndex(DockableGroup::DOCKED_BOTTOM, this));
	o = new Item("Dock &Left");
	o->callback(menu_cb, new MenuIndex(DockableGroup::DOCKED_LEFT, this));
	o = new Item("Dock &Right");
	o->callback(menu_cb, new MenuIndex(DockableGroup::DOCKED_RIGHT, this));
	
	new Divider();
	o = new Item("&Vertical");
	o->callback(menu_cb, new MenuIndex(VERTICAL_MENUINDEX, this));
	o = new Item("&Horizontal");
	o->callback(menu_cb, new MenuIndex(HORIZONTAL_MENUINDEX, this));
	o = new Item("&Float");
	o->callback(menu_cb, new MenuIndex(FLOAT_MENUINDEX, this));
	new Divider();
	o = new Item("&Hide");
	o->callback(menu_cb, new MenuIndex(HIDE_MENUINDEX, this));
	end();
}

void tgGripper::menu_cb(Widget* o, void* p) {
	MenuIndex *mi = (MenuIndex*)p;
	DockableGroup *dg = dynamic_cast<DockableGroup*>(mi->gripper_->parent());
	
	switch(mi->index_) {
		case DockableGroup::DOCKED_TOP:
		case DockableGroup::DOCKED_BOTTOM:
		case DockableGroup::DOCKED_LEFT:
		case DockableGroup::DOCKED_RIGHT:
			dg->state(dg->region(0)->dock(dg, mi->index_));
			break;
		case VERTICAL_MENUINDEX:
			dg->orientation(DockableGroup::ORIENTED_VERTICAL);
			break;
		case HORIZONTAL_MENUINDEX:
			dg->orientation(DockableGroup::ORIENTED_HORIZONTAL);
			break;
		case FLOAT_MENUINDEX:
			if(dg->state() != DockableGroup::DOCKED_NONE) {
				dg->state(dg->region(0)->undock(dg, dg->state()));
			}
			break;
		case HIDE_MENUINDEX:
			dg->window()->hide();
	}
}

void tgGripper::draw() {
	PopupMenu::draw();
}

int tgGripper::handle(int e) {
	if(e == PUSH && event_button() == RightButton) {
		popup();
	}
	return 1;
}

tgChildInfo::tgChildInfo(Widget *o, uchar f) {
	widget_ = o;
	flag_ = f;
	iw_ = o->w();
	ih_ = o->h();
}

void ToolGroup::insert(tgChildInfo *o, int index) {
	if (lcinfo_ == 0) {
		cinfo_ = new tgChildInfo*[1];
		cinfo_[0] = o;
	} else {
		if (!(lcinfo_ & (lcinfo_-1))) {
			tgChildInfo** newarray = new tgChildInfo*[2*lcinfo_];
			memcpy(newarray, cinfo_, lcinfo_*sizeof(tgChildInfo*));
			delete[] cinfo_;
			cinfo_ = newarray;
		}
		for (int j = lcinfo_; j > index; --j) cinfo_[j] = cinfo_[j-1];
		cinfo_[index] = o;
	}
	++lcinfo_;
}

ToolGroup::ToolGroup(int X, int Y, int W, int H, const char *l) : DockableGroup (0, 0, W, H, l) {
	lcinfo_ = 0;
	gripper_ = new tgGripper();
	child_info(tgChildInfo::INTERNAL_CHILD);
	
	window()->clear_border();
	box(ENGRAVED_BOX);
}

ToolGroup::~ ToolGroup() {
}

void ToolGroup::child_info(int f) {
	if(!children()) return;
	add(new tgChildInfo(child(children() - 1), f));
}

void ToolGroup::init_child() {
	if(lcinfo_ < children()) {
		int i;
		for(i = 0; i < lcinfo_; i++) {
			if(cinfo_[i]->widget_ != child(i)) {
				insert(new tgChildInfo(child(i)), i);
			}
		}
		while(i < children()) {
			add(new tgChildInfo(child(i)));
			i++;
		}
	}
}

void ToolGroup::update_all() {
	init_child();
	
_AGAIN_:
	int tx, ty;
	Rectangle r(w(), h());
	box()->inset(r);
	
	int cpos;
	
	if(orientation()) {
		tx = r.x() + GRIPPER_GAP + gripper_->thickness() + GRIPPER_GAP;
		ty = r.y();
		gripper_->resize(r.x() + GRIPPER_GAP, r.y() + GRIPPER_GAP,
										 gripper_->thickness(), r.h() - (GRIPPER_GAP << 1));
		cpos = tx;
	}
	else {
		tx = r.x();
		ty = r.y() + GRIPPER_GAP + gripper_->thickness() + GRIPPER_GAP;
		gripper_->resize(r.x() + GRIPPER_GAP, r.y() + GRIPPER_GAP,
										 r.w() - (GRIPPER_GAP << 1), gripper_->thickness());
		cpos = ty;
	}
	int maxw = w(), maxh = h();
	int cchild = children();
	
	for (int i = 0; i < cchild; i++) {
		Widget* o = child(i);
		if(cinfo_[i]->flag_ & tgChildInfo::INTERNAL_CHILD) continue;
		
		int X,Y,W,H;
		if (orientation()) {
			X = cpos;
			W = cinfo_[i]->iw();
			if(cinfo_[i]->flag_ & tgChildInfo::FIXED_HEIGHT) {
				maxh = (maxh < cinfo_[i]->ih())? cinfo_[i]->ih() : maxh;
				Y = ty + (r.h() - cinfo_[i]->ih()) / 2;
				H = cinfo_[i]->ih();
			}
			else {
				Y = ty;
				H = r.h();
			}
			cpos += W;
		} 
		else {
			if(cinfo_[i]->flag_ & tgChildInfo::FIXED_WIDTH) {
				maxw = (maxw < cinfo_[i]->iw())? cinfo_[i]->iw() : maxw;
				X = tx + (r.w() - cinfo_[i]->iw()) / 2;
				W = cinfo_[i]->iw();
			}
			else {
				X = tx;
				W = r.w();
			}
			Y = cpos;
			H = cinfo_[i]->ih();
			cpos += H;
		}
		if (X != o->x() || Y != o->y() || W != o->w() || H != o->h()) {
			o->resize(X,Y,W,H);
		}
	}
	if(w() != maxw || h() != maxh) {
		resize(x(), y(), maxw, maxh);
		goto _AGAIN_;
	}
}

void ToolGroup::layout() {
	update_all();
}

int ToolGroup::handle(int event) {
	static uchar flag_;
	static int oldx_, oldy_;

	switch(event) {
		case PUSH:
			if(event_inside(*gripper_) && event_button() == LeftButton) {
				oldx_ = event_x();
				oldy_ = event_y();
				flag_ = 1;
				if(state() != DOCKED_NONE) 
					state(region(0)->check_undock(this, state()));
				return 1;
			}
			break;
		case RELEASE:
			flag_ = 0;
			if(!(state() != DOCKED_NONE)) state(region(0)->check_dock(this));
			else region(0)->docked_layout(state());
			break;
			case DRAG: //for smooth drag this cannot be in grip handle
				if(flag_) {
					if(state() != DOCKED_NONE) {
						position(x() + event_x() - oldx_, y() + event_y() - oldy_);
						state(region(0)->check_undock(this, state()));
					}
					else {
						window()->position(window()->x() + event_x() - oldx_, window()->y() + event_y() - oldy_);
					}
					return 1;
				}
				break;
	}
	return Group::handle(event);
}

//
// End of "$Id ToolGroup.cxx Thu Feb 7 2006 z_hossain $".
//
